local _G = _G

-- addon name and namespace
local ADDON, NS = ...

-- local functions
local strlower          = strlower
local strfind           = strfind
local strupper          = strupper
local string_format     = string.format
local tinsert           = table.insert
local tremove           = table.remove
local table_sort        = table.sort
local ipairs            = ipairs
local pairs             = pairs
local time              = time
local floor             = floor
local gsub              = gsub

local GetNumGroupMembers    = _G.GetNumGroupMembers
local GetNumSubgroupMembers = _G.GetNumSubgroupMembers
local InCombatLockdown      = _G.InCombatLockdown
local IsAltKeyDown          = _G.IsAltKeyDown
local IsControlKeyDown      = _G.IsControlKeyDown
local IsInRaid              = _G.IsInRaid
local IsShiftKeyDown        = _G.IsShiftKeyDown
local RequestRaidInfo       = _G.RequestRaidInfo
local UnitName              = _G.UnitName

local LE_PARTY_CATEGORY_HOME = _G.LE_PARTY_CATEGORY_HOME

-- aux variables
local _

local characters = {}
local instances  = {}

-- setup libs
local LibStub   = LibStub

-- toast notifications
local LibToast = LibStub("LibToast-1.0")

-- get translations
local L         = LibStub:GetLibrary("AceLocale-3.0"):GetLocale(ADDON)

-- addon
local Addon = LibStub:GetLibrary("AceAddon-3.0"):NewAddon(ADDON, "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")

-- internal event handling
Addon.callbacks = LibStub("CallbackHandler-1.0"):New(Addon)

-- constants
Addon.MODNAME   = "Argus"
Addon.FULLNAME  = "Argus"
Addon.SHORTNAME = "Argus"

-- modules
local Options       = nil
local Interface     = nil
local Tooltip       = nil
local MinimapButton = nil
local DataBroker    = nil

-- icons
local ICON          = "Interface\\Addons\\"..ADDON.."\\icon.tga"
local ICON_DISABLED = "Interface\\Addons\\"..ADDON.."\\icondisabled.tga"

LibStub:GetLibrary('LibWho-2.0'):Embed(Addon)

local MIN_REQUEST_INTERVAL = 10
local CYCLES_UNTIL_TIMEOUT = 3

-- close toast callback
local function CloseToast()
	Addon:TriggerAction("show")
end

-- infrastcructure
function Addon:OnInitialize()
	-- addon constants
	self.MAX_TIME_RANGE = 120	

	-- addon variables
	
	-- aux vars 
	self.playerSavedRaids = {}
	self.notificationSent = {}
	
	-- who request handling
	self.requestQueue         = NS:CreateFIFOQueue()
	self.requestInstanceTimer = nil
	self.requestCycleTimer    = nil
	self.requestReady         = false
	self.requestActive        = false
	self.elapsedCount         = 0
	
	-- monitored instances
	self.monitored = {}
	
	-- data
	self.characterData = {}
	self.instanceData  = {}
	
	-- track combat 
	self.inCombat = false
	
	-- debugging
	self.debug = false
	
	-- unit information
	self.raidNames   = {}
	self.PlayerName  = UnitName("player")
	self.playerName  = strlower(self.PlayerName)

	self.classCache  = {}
	
	-- offset to local time
	local gHour = GetGameTime()
	self.timeOffset = (tonumber(gHour) - date("!%H")) * 3600
	
	-- processed data
	self.taintedData       = true
	self.matchedCharacters = 0
	self.matchedInstances  = 0
	
	self:RegisterChatCommand("argus", "ChatCommand")
	
	-- setup toast support
	LibToast:Register(Addon.FULLNAME, function(toast, ...)
		toast:SetTitle(Addon.FULLNAME)
		toast:SetIconTexture(ICON)
		toast:SetText(...)
		toast:SetUrgencyLevel("moderate")
		toast:SetPrimaryCallback(_G.SHOW, CloseToast)
	end)	
end

function Addon:OnEnable()
	-- set module references
	Options       = self:GetModule("Options")
	Interface     = self:GetModule("Interface")
	Tooltip       = self:GetModule("Tooltip")
	MinimapButton = self:GetModule("MinimapButton")
	DataBroker    = self:GetModule("DataBroker")

	-- redo: offset to local time
	-- seems on initialize the game time is not always set properly
	local gHour = GetGameTime()
	self.timeOffset = (tonumber(gHour) - date("!%H")) * 3600
	
	self:SetupEventHandlers()
	
	-- update keyword data
	self:InitInstanceData()
	
	-- update cached unit data
	self:UpdateRaidUnits()

	-- set initial combat state
	self:UpdateCombatState()
	
	MinimapButton:SetShow(Options:GetSetting("Minimap"))

	self:UpdateAllMonitoredInstances()
	self:UpdateMonitoring()

	self:UpdateIcon()
	self:Update()
	
	-- request info about saved instances
	RequestRaidInfo() 
end

function Addon:OnDisable()
	self:ResetEventHandlers()
		
	self:ClearRequests()

	MinimapButton:SetShow(false)
end

function Addon:OnOptionsReloaded()
	MinimapButton:SetShow(Options:GetSetting("Minimap"))

	self:ClearRequests()
	self:UpdateAllMonitoredInstances()
	self:UpdateMonitoring()
	
	self.taintedData = true

	self:UpdateIcon()
	self:Update()
end

function Addon:ChatCommand(input)
    if input then  
		args = NS:GetArgs(input)
		
		self:TriggerAction(unpack(args))
	        
        NS:ReleaseTable(args)
	else
		self:TriggerAction("help")
	end
end

function Addon:OnClick(button)
	if button == "RightButton" then 
		if IsShiftKeyDown() then
			-- unused
		elseif IsControlKeyDown() then
			-- unused
		elseif IsAltKeyDown() then
			-- unused
		else
			-- open options menu
			Addon:TriggerAction("menu")
		end
	elseif button == "LeftButton" then 
		if IsShiftKeyDown() then
			-- unused
		elseif IsControlKeyDown() then
			Addon:TriggerAction("request")
		elseif IsAltKeyDown() then
			-- toggle monitoring
			Addon:TriggerAction("toggle")
		else
			-- show finder log
			Addon:TriggerAction("show")
		end
	end
end

function Addon:TriggerAction(action, ...)
    local args = NS:NewTable(...)

	action = type(action) == "string" and string.lower(action) or ""	

	if action == "menu" then
		-- open options menu
		InterfaceOptionsFrame_OpenToCategory(self.FULLNAME)
	elseif action == "version" then
		-- print version information
		self:PrintVersionInfo()
	elseif action == "debug" then
		if args[1] == "on" then
			self:Output("debug mode turned on")
			self.debug = true
		end
		if args[1] == "off" then
			self:Output("debug mode turned off")
			self.debug = false
		end
	elseif action == "toggle" then
		Options:ToggleSetting("MonitoringActive")
		self:Output(L["Monitoring is "] .. (Options:GetSetting("MonitoringActive") and L["on"] or L["off"]))
	elseif action == "on" then
		Options:SetSetting("MonitoringActive", true)
		self:Output(L["Monitoring is "] .. (Options:GetSetting("MonitoringActive") and L["on"] or L["off"]))
	elseif action == "off" then
		Options:SetSetting("MonitoringActive", false)
		self:Output(L["Monitoring is "] .. (Options:GetSetting("MonitoringActive") and L["on"] or L["off"]))
	elseif action == "show" then
		Interface:ShowLogWindow()
	elseif action == "saved" then
		if next(self.playerSavedRaids) then
			for instance, _ in pairs(self.playerSavedRaids) do
				self:Output(instance)
			end
		else
			self:Output("No saved raids.")
		end
	elseif action == "info" then
		local instance = args[1]
		self:Output("Info for instance: " .. tostring(instance))
		self:Output("Valid - " .. tostring(NS:IsValidInstance(instance)))
		self:Output("Localized - " .. tostring(NS:TranslateInstance(instance)))
		self:Output("Monitored - " .. tostring(self:IsMonitored(instance)))
		self:Output("Activated - " .. tostring(Options:IsMonitored(instance)))
		self:Output("Saved - " .. tostring(self:IsSaved(instance)))
		self:Output("Excluded as saved - " .. tostring(self:ExcludeAsSaved(instance)))
		self:Output("Not monitored as saved - " .. tostring(self:NotMonitoredDueToLock(instance)))
	elseif action == "status" then
		self:Output("Queue length: " .. tostring(self.requestQueue:Count()))
		self:Output("Queue first: " .. tostring(self.requestQueue:Peek()))
		self:Output("Active: " .. tostring(self.requestActive))
		self:Output("Ready: " .. tostring(self.requestReady))
	elseif action == "request" then
		self:RequestMonitored()
	else -- if action == "help" then
		-- display help
		self:Output(L["Usage:"])
		self:Output(L["/braidfinder arg"])
		self:Output(L["/brfind arg"])
		self:Output(L["Args:"])
		self:Output(L["version - display version information"])
		self:Output(L["menu - display options menu"])
		self:Output(L["on - activate monitoring"])
		self:Output(L["off - deactivate monitoring"])
		self:Output(L["show - show log window"])
		self:Output(L["help - display this help"])
	end
    
    NS:ReleaseTable(args)
end

function Addon:CreateTooltip(obj, autoHideDelay)
	Tooltip:Create(obj, autoHideDelay)
end

function Addon:RemoveTooltip()
	Tooltip:Remove()
end

function Addon:InitInstanceData()
	local now = time()
	
	-- for all extensions
	for extension, instances in ipairs(NS.instances) do		
		-- for all instances
		for _, instance in ipairs(instances) do
			if not self.instanceData[instance] then
				local data = {
					instance   = instance,
					timestamp  = now,
					count      = 0,
					characters = {},
				}
				
				self.instanceData[instance] = data
			end

			self:UpdateMonitoredInstance(nil, instance, true)
		end
	end
end

function Addon:SetupEventHandlers()
	Options.RegisterCallback(self, ADDON .. "_MONITORING_CHANGED", "UpdateMonitoredInstance")	
	Options.RegisterCallback(self, ADDON .. "_SETTING_CHANGED",    "UpdateSetting")	

	-- register raid info events
	self:RegisterEvent("UPDATE_INSTANCE_INFO")
	self:RegisterEvent("RAID_INSTANCE_WELCOME")
	self:RegisterEvent("CHAT_MSG_SYSTEM")
	
	-- register raid unit events
	self:RegisterEvent("PARTY_CONVERTED_TO_RAID", "UpdateRaidUnits")
	self:RegisterEvent("GROUP_ROSTER_UPDATE", "UpdateRaidUnits")

	-- register events to track combat
	self:RegisterEvent("PLAYER_REGEN_ENABLED",  "UpdateCombatState", false)
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "UpdateCombatState", true)
end

function Addon:ResetEventHandlers()
	Options.UnregisterCallback(self, ADDON .. "_MONITORING_CHANGED")	
	Options.UnregisterCallback(self, ADDON .. "_SETTING_CHANGED")

	-- unregister chat events
	self:UnregisterEvent("CHAT_MSG_CHANNEL")
	self:UnregisterEvent("CHAT_MSG_GUILD")
	self:UnregisterEvent("CHAT_MSG_OFFICER")	
	
	-- unregister raid info events
	self:UnregisterEvent("UPDATE_INSTANCE_INFO")
	self:UnregisterEvent("RAID_INSTANCE_WELCOME")
	self:UnregisterEvent("CHAT_MSG_SYSTEM")

	-- unregister raid unit events
	self:UnregisterEvent("PARTY_CONVERTED_TO_RAID")
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED")
	
	-- register events to track combat
	self:UnregisterEvent("PLAYER_REGEN_ENABLED")
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
end

-- update functions
function Addon:Update()
	self:UpdateData()
	self:UpdateLabel()
end

-- aggregate some meaningful data from the list of matches for display on label/tooltip and internal processing
-- matchedCharacters/Instances:
-- 		aggregates the total number of characters/instances from valid matches 
function Addon:UpdateData()
	Interface:UpdateLogList()

	if self.taintedData == false then
		return
	end

	NS:ClearTable(characters)
	NS:ClearTable(instances)
	
	for name, data in self:IterateCharacterData() do
		if self:IsMonitored(data.instance) then
			characters[name] = true
			instances[data.instance] = true
		end
	end
	
	self.matchedCharacters = 0
	
	for _ in pairs(characters) do
		self.matchedCharacters = self.matchedCharacters + 1
	end

	self.matchedInstances = 0
	
	for _ in pairs(instances) do
		self.matchedInstances = self.matchedInstances + 1
	end	

	self.taintedData = false
end

function Addon:UpdateLabel()
	local text = L["I: "] .. tostring(self.matchedInstances) .. " / " .. L["P: "] .. tostring(self.matchedCharacters)
	
	if not Options:GetSetting("MonitoringActive") then
		text = NS:Colorize("GrayOut", text)
	elseif not self:IsAnyInstanceMonitored() then
		text = NS:Colorize("Yellow", text)
	elseif self.matchedInstances == 0 then
		text = NS:Colorize("Red", text)
	else
		text = NS:Colorize("Green", text)
	end
			
	DataBroker:SetText(text)
end

function Addon:UpdateIcon()
	local icon = self:GetIcon()
	
	DataBroker:SetIcon(icon)
	MinimapButton:SetIcon(icon)
end

function Addon:GetIcon()
	if Options and Options:GetSetting("MonitoringActive") then
		return ICON
	else
		return ICON_DISABLED
	end
end

-- settings
function Addon:UpdateSetting(event, setting, value, old)
	if setting == "MonitoringActive" then
		self:UpdateMonitoring()
		self:UpdateIcon()
		self:Update()	
	elseif setting == "ExcludeSavedRaids" then
		self:UpdateAllMonitoredInstances()
	elseif setting == "LookupInterval" then
		self:UpdateIntervalTimer()
	elseif setting == "Minimap" then
		MinimapButton:SetShow(Options:GetSetting("Minimap"))
	end
end

function Addon:GetSetting(setting)
	return Options:GetSetting(setting)
end

-- instance handling
function Addon:UpdateAllMonitoredInstances()
	for instance in NS:IterateInstances() do
		self:UpdateMonitoredInstance(nil, instance, true)
	end
	
	self.taintedData = true
	
	self:Update()
end

function Addon:UpdateMonitoredInstance(event, instance, silent)
	if not NS:IsValidInstance(instance) then
		return
	end
	
	local monitored = Options:IsMonitored(instance) and not self:ExcludeAsSaved(instance)
	
	if monitored then
		self.monitored[instance] = true
		
		self:EnqueueRequest(instance)
	else
		self.monitored[instance] = nil
	end
	
	self.taintedData = true
	
	if not silent then
		self:Update()	
	end
end

function Addon:IsMonitored(instance)
	return self.monitored[instance] == true
end

function Addon:IsAnyInstanceMonitored()
	return next(self.monitored) ~= nil
end

function Addon:IterateMonitoredInstances()
	return pairs(self.monitored)
end

function Addon:ToggleMonitored(instance)
	Options:ToggleMonitored(instance)
end

function Addon:NotMonitoredDueToLock(instance)
	return Options:IsMonitored(instance) and self:ExcludeAsSaved(instance)
end

-- update cached data function
function Addon:UpdateRaidInfo()
	NS:ClearTable(self.playerSavedRaids)

	local numSaved = GetNumSavedInstances()
	
	-- we do not distinguish between different difficulties
	for i = 1, numSaved do
		local instanceName, instanceID, instanceReset, instanceDifficulty, locked, extended, instanceIDMostSig, isRaid, maxPlayers, difficultyName, maxBosses, defeatedBosses = GetSavedInstanceInfo(i)
		
		if isRaid and locked then
			instance = self:FuzzySearchInstance(instanceName)

			if instance then
				self.playerSavedRaids[instance] = true
			end
		end
	end
end

-- since we need to map the localized name returned by GetSavedInstanceInfo to our internal non-localized instance name
-- we perform a very simple matching algorithm to do the task
-- if pattern contains all of the significant words (longer than 3 chars) we assume we found the match
function Addon:FuzzySearchInstance(pattern)
	if not pattern then
		return
	end
	
	pattern = strlower(pattern)

	for instance in NS:IterateInstances() do
		local localized = NS:TranslateInstance(instance)
		localized = strlower(localized)
		
		local matched = true
		
		for word in string.gmatch(localized, "[^%s%p%c]+") do
		    -- significant length of a word is 4 characters
			-- not sure if that works well with all possible localizations
			if string.len(word) > 3 then
				matched = matched and strfind(pattern, word)
			end
		end
		
		if matched then
			return instance
		end
	end
	
	return 
end

function Addon:UpdateRaidUnits()
	NS:ClearTable(self.raidNames)
	
	if GetNumGroupMembers(LE_PARTY_CATEGORY_HOME) > 0 then
		if IsInRaid() then
			for i = 1, 40, 1 do
				local unit = "raid"..i
				local name = UnitName(unit)
				
				if name then
					self.raidNames[name] = true
				end
			end
		else
			for i = 1, 4, 1 do
				local unit = "party"..i
				local name = UnitName(unit)
				
				if name then
					self.raidNames[name] = true
				end
			end
		end
	end
end

function Addon:UpdateCombatState(inCombat)
	if inCombat == nil then
		self.inCombat = InCombatLockdown()
	else
		self.inCombat = inCombat
	end
end

function Addon:InCombat()
	return self.inCombat
end

-- event handlers
function Addon:UPDATE_INSTANCE_INFO()
	self:UpdateRaidInfo()
end

function Addon:RAID_INSTANCE_WELCOME()
	RequestRaidInfo()
end

function Addon:CHAT_MSG_SYSTEM(msg)
	if tostring(msg) == _G["INSTANCE_SAVED"] then
		RequestRaidInfo()
	end
end

-- user functions
function Addon:PrintVersionInfo()
    self:Output(L["Version"] .. " " .. NS:Colorize("White", GetAddOnMetadata(ADDON, "Version")))
end

-- utilities
function Addon:Output(msg)
	if ( msg ~= nil and DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( self.MODNAME..": "..msg, 0.6, 1.0, 1.0 )
	end
end

function Addon:CacheCharacterClass(name, class)
	if not name or not class or not _G["RAID_CLASS_COLORS"][class] then
		return
	end
	
	if not self.classCache[name] then
		self.classCache[name] = class
	end
end

function Addon:ColorizeChar(name)
	if not name then
		return
	end
	
	local class = self:GetClass(name)
	
	if not class then
		local result = self:UserInfo(name, {queue = self.WHOLIB_QUEUE_QUIET, timeout = -1, callback = 'UnitInfoCallback'}) --flags = self.WHOLIB_FLAG_ALWAYS_CALLBACK, 
		
		if result then
			self:UnitInfoCallback(result)
			
			class = self.classCache[name]
		end
	end
	
	if class then
		local colortable = _G["CUSTOM_CLASS_COLORS"] or _G["RAID_CLASS_COLORS"]
		local color = colortable[class]
		return "|cff" .. string.format("%02x%02x%02x", color.r*255, color.g*255, color.b*255) .. name .. "|r"
	else
		return NS:Colorize("GrayOut", name)
	end
end

function Addon:GetClass(name)
	return self.classCache[name]
end

function Addon:UnitInfoCallback(user, time)
	if user and user.Class then
		self:CacheCharacterClass(user.Name, gsub(strupper(user.Class), " ", ""))
	end
end

function Addon:IsMonitored(instance)
	return self.monitored[instance] == true
end

function Addon:IsSaved(instance)
	return self.playerSavedRaids[instance] == true
end

function Addon:ExcludeAsSaved(instance)
	return Options:GetSetting("ExcludeSavedRaids") and self:IsSaved(instance)
end

function Addon:GetCharacterCount(instance)
	return self.instanceData[instance] and self.instanceData[instance].count or 0
end

function Addon:PlaySoundFile(sound)
	if type(sound) ~= "string" then
		return
	end

	if ((string.sub(sound, -4) == ".ogg") or (string.sub(sound, -4) == ".mp3")) then
		PlaySoundFile(sound)
	else
		PlaySound(sound)
	end
end

function Addon:IsInRaid(name)
	return self.raidNames[name] and true
end

function Addon:WhisperCharacter(character)
	if not character then
		return
	end

	DEFAULT_CHAT_FRAME.editBox:Show()
	DEFAULT_CHAT_FRAME.editBox:SetText("/whisper " .. character .. " ")
end

function Addon:GetTimeOffset()
	return self.timeOffset
end

-- testing
function Addon:Debug(msg)
	if self.debug then
		if ( msg ~= nil and DEFAULT_CHAT_FRAME ) then
			DEFAULT_CHAT_FRAME:AddMessage( self.MODNAME .. " (dbg): " .. msg, 1.0, 0.37, 0.37 )
		end
	end
end

-- data handling
function Addon:UpdateInstanceData(instance, characters)
	if not self:IsMonitored(instance) then
		return
	end
	
	characters = characters or {}
	
	local now = time()
	
	local instanceData = self.instanceData[instance]
	local characterData   = self.characterData
	
	local oldCount = instanceData.count
	
	local changed = false
	
	local localized = NS:TranslateInstance(instance)
	
	-- throw out characters no longer present
	for name in pairs(instanceData.characters) do
		if not characters[name] then
			instanceData.characters[name] = nil
			instanceData.count = instanceData.count - 1
			
			-- remove character without instance
			characterData[name] = nil
			
			changed = true
		end
	end

	-- add new characters
	for name, character in pairs(characters) do
		if not instanceData.characters[name] then
			instanceData.characters[name] = true
			instanceData.count = instanceData.count + 1
			
			if not characterData[name] then
				characterData[name] = character
			
				self:CacheCharacterClass(character.name, character.class)
			else
				-- remove char from other instance
				local lastInstance = self.instanceData[characterData[name].instance]
				
				if lastInstance then					
					lastInstance.characters[name] = nil
					lastInstance.count = instanceData.count - 1
				end
			end
			
			character.instance  = instance
			character.localized = localized
			
			changed = true
		end
		
		characterData[name].timestamp = now
	end

	instanceData.timestamp = now
	
	self.taintedData = true	
	self:Update()
		
	if instanceData.count ~= oldCount and oldCount == 0 then
		self:Notify(instance)
	end
end					

function Addon:Notify(instance)
	if not self:IsMonitored(instance) then
		return
	end	
	
	local data = self.instanceData[instance]
	
	local count = data.count or 0
	
	if count == 0 then
		-- reset notification
		self.notificationSent[instance] = nil
		
		return
	elseif self.notificationSent[instance] then
		-- already sent
		return
	end
	
	if Options:GetSetting("NotifyText") then
		self:Output(string.format(L["Found %s active characters in %s."], tostring(count), instance))
		self.notificationSent[instance] = true
	end

	if Options:GetSetting("NotifySound") then
		self:PlaySoundFile(Options:GetSoundFile(Options:GetSetting("NotificationSound")))
		self.notificationSent[instance] = true
	end
	
	if Options:GetSetting("NotifyToast") then
		LibToast:Spawn(Addon.FULLNAME, string.format(L["Found %s active characters in %s."], tostring(count), NS:Colorize("Green", instance) .. "."))
		self.notificationSent[instance] = true
	end	
end

function Addon:IterateCharacterData()
	return pairs(self.characterData)
end

function Addon:IterateInstanceData()
	return pairs(self.instanceData)
end

function Addon:GetInstanceData(instance)
	return self.instanceData[instance]
end

-- WhoLib request handling
function Addon:UpdateMonitoring()
	if self:IsEnabled() and Options:GetSetting("MonitoringActive") then
		self:RequestMonitored()
	else
		self:ClearRequests()
	end
	
	self:UpdateIntervalTimer()
end

function Addon:UpdateIntervalTimer()
	local interval = Options:GetSetting("LookupInterval") * 60
	
	if self:IsEnabled() and Options:GetSetting("MonitoringActive") and interval > 0 then
		-- interval option changed, rescheduler timer
		if self.requestCycleTimer and interval ~= self.requestCycleTimer.delay then
			self:CancelTimer(self.requestCycleTimer)
			self.requestCycleTimer = nil
		end
		
		if not self.requestCycleTimer then
			self.requestCycleTimer = self:ScheduleRepeatingTimer("RequestMonitored", interval)
		end
	else
		if self.requestCycleTimer then
			self:CancelTimer(self.requestCycleTimer)
			self.requestCycleTimer = nil
		end
	end

	if self:IsEnabled() then
		if not self.requestInstanceTimer then
			self.requestInstanceTimer = self:ScheduleRepeatingTimer("RequestInstanceTimerElapsed", MIN_REQUEST_INTERVAL)
			
			self.requestReady = true
		end
	else
		if self.requestInstanceTimer then
			self:CancelTimer(self.requestInstanceTimer)
			self.requestInstanceTimer = nil
			
			self.requestReady = false			
		end
	end
end

function Addon:RequestMonitored()
	if not self:IsEnabled() then
		return
	end
	
	for instance in self:IterateMonitoredInstances() do
		self:EnqueueRequest(instance)
	end
end

function Addon:EnqueueRequest(instance)
	if not self:IsEnabled() then
		return
	end
	
	if not self:IsMonitored(instance) then
		return
	end

	if self.requestQueue:Contains(instance) then
		return 
	end
	
	self.requestQueue:Enqueue(instance)
	
	self:TryExecuteNextRequest()
end

function Addon:ClearRequests()
	self.requestQueue:Clear()
end

function Addon:TryExecuteNextRequest()
	if not self:IsEnabled() then
		return
	end

	if not self.requestReady or self.requestActive then
		return
	end
	
	local instance = self.requestQueue:Peek()
	
	while instance and not self:IsMonitored(instance) do
		self.requestQueue:Dequeue()
		instance = self.requestQueue:Peek()
	end
	
	if not instance then
		return
	end

	local localizedInstance = NS:TranslateInstance(instance)	

	-- z- option in who search that restricts search to zones doesn't work if zone starts with 'The '
	local searchString = localizedInstance.gsub(localizedInstance, "^The ", "")
	
	-- execute request
	-- NOTE: since 5.3 z- option doesn't seem to work with instances anymore :-/
	-- self:Who("z-" .. searchString, {queue = self.WHOLIB_QUEUE_SCANNING, timeout = 1, callback = 'WhoLibResultCallback'})
	self:Who(searchString, {queue = self.WHOLIB_QUEUE_SCANNING, timeout = 1, callback = 'WhoLibResultCallback'})
	
	self.requestActive = true
	self.requestReady  = false
end

function Addon:RequestInstanceTimerElapsed()
	self.requestReady = true
	
	-- if timer elapsed more often than the specified number for timeout consider the request s failed and reset activity flag
	-- current request will be sent again
	self.elapsedCount = self.elapsedCount + 1

	if self.elapsedCount > CYCLES_UNTIL_TIMEOUT then
		self.elapsedCount = 0
		self.requestActive = false
	end
	
	self:TryExecuteNextRequest()
end

-- Event: WHOLIB_QUERY_RESULT - query, results, complete, name
-- Args
-- query
    -- string - search string 
-- results
    -- table - table of results 
-- results[i].Name
    -- string - name of the player 
-- results[i].Online
    -- true if online, false if offline, nil if unknown (more results than could be displayed) 
-- results[i].Guild
    -- string - guild or '' 
-- results[i].Class
    -- string - class 
-- results[i].Race
    -- string - race 
-- results[i].Level
    -- string - level 
-- results[i].Zone
    -- string - zone 
-- complete
    -- boolean - shows whether all results could be returned (true) or not (false), if not, do a more specific query 
-- name
    -- string - if the query was initiated by a :UserInfo() call, then this is the player name of the :UserInfo() call, otherwise nil 
function Addon:WhoLibResultCallback(query, results, complete)
	if not results then
		return
	end
	
	local instance = self.requestQueue:Peek()
	
	local localizedInstance = NS:TranslateInstance(instance)	

	-- z- option in who search that restricts search to zones doesn't work if zone starts with 'The '
	local searchString = localizedInstance.gsub(localizedInstance, "^The ", "")
		
	self:Debug("WhoLibResultCallback: query " .. tostring(query) .. ", " .. tostring(instance) .. ", " .. tostring(localizedInstance))
	self:Debug("WhoLibResultCallback: results " .. tostring(results) .. ", complete " .. tostring(complete))
	
	if query ~= searchString then
		return
	end
	
	if self:IsMonitored(instance) then	
		self:Debug("WhoLibResultCallback: is monitored instance " .. tostring(results and #results or 0	))
		-- handle result data
		local characters = {}

		for _, result in ipairs(results) do
			self:Debug("WhoLibResultCallback: char " .. tostring(result.Name))

			if result.Zone == localizedInstance then
				self:Debug("WhoLibResultCallback: in correct zone")
				characters[result.Name] = {
					name  = result.Name,
					class = result.Class,
					guild = result.Guild,
					level = result.Level,
					race  = result.Race,
					zone  = result.Zone,
				}				
			end			
		end
		
		self:UpdateInstanceData(instance, characters)
	end
		
	-- clear instance from queue and clear active flag
	self.requestQueue:Dequeue()
	self.requestActive = false
	self.elapsedCount = 0
	
	self:TryExecuteNextRequest()
end
